;===============================================================================
;
;libstr30c: CHARACTER STRING MANIPULATION LIBRARY V3
;
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;*                                                                             *
;* Copyright 2015 by William J Brier                                          *
;*                                                                             *
;* Redistribution & use of this software in source and/or binary forms, with   *
;* or without modification, are permitted, provided that the following condi-  *
;* tions are met:                                                              *
;*                                                                             *
;* 1) Redistributions of source code must retain the above copyright notice,   *
;*    this list of conditions & the below disclaimer.                          *
;*                                                                             *
;* 2) Redistribution in binary form must reproduce the above copyright notice, *
;*    this list of conditions & the below disclaimer in the documentation &    *
;*    other materials included with the distribution.                          *
;*                                                                             *
;* 3) The names of the copyright holder and/or its contributors may not be     *
;*    used to endorse or promote products derived from this software without   *
;*    specific prior written permission.                                       *
;*                                                                             *
;* DISCLAIMER                                                                  *
;*                                                                   *
;* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS & CONTRIBUTORS AS IS.  *
;* Any express or implied warranties, including, but not limited to, the imp-  *
;* lied warranties of merchantability & fitness for a particular purpose are   *
;* disclaimed.  In no event shall the copyright owner and/or contributors be   *
;* liable for any direct, indirect, incidental, special, exemplary, or conse-  *
;* quential damages (including, but not limited to, procurement of substitute  *
;* goods or services; loss of use, data, or profits; or business interruption) *
;* however caused & on any theory of liability, whether in contract, strict    *
;* liability, or tort (including negligence or otherwise) arising in any way   *
;* out of the use of this software, even if advised of the possibility of such *
;* damage.                                                                     *
;*                                                                             *
;* If any provision set forth herein is not acceptable to you, DO NOT USE THIS *
;* SOFTWARE & immediately delete it from your system.                          *
;*                                                                             *
;* Refer to commentary at the end of this file for more information.           *
;*                                                                             *
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;
;===============================================================================
;
;	.org $000000
;
;_STRLIB_		;uncomment to enable "far" calls
;
;	Refer to documentation for proper usage of above statements.
;
;===============================================================================
;
_strlen_	.macro .s
         ldx .s
         ldy .s+2
         per *+5
         brl __strlen__
	.endm
;
;===============================================================================
;
;_strlen_: DETERMINE siz(S)
;
;	
;	This function is an internal call & must
;	not be invoked by a user programs.
;	
;
__strlen__ php
           phd
;
;
;
;LOCAL DEFINTIONS
;
.s_word	=2
.s_wsf	=4
.seta	=%00100000
.setx	=%00010000
.setr	=.seta|.setx
.sr_car	=%00000001
.sr_bdm	=%00001000
.sptr	=0
;
;
         rep #.setr|.sr_bdm
         sec
         tsc
         sbc !#.s_wsf
         tcs
         inc
         tcd
         stx .sptr
         sty .sptr+.s_word
         sep #.seta
         ldy !#0
;
.main    lda [.sptr],y
         beq .done
;
         iny
         bpl .main
;
.done    rep #.seta|.sr_car
         tsc
         adc !#.s_wsf
         tcs
         pld
         plp
         phy
         ply
         rts
;
	.if .ref(strcat)
;===============================================================================
;
;strcat: CONCATENATE STRING S2 to STRING S1
;
;	
;	Invocation example: PEA #S2_PTR >> 16   ;*S2 MSW
;	                    PEA #S2_PTR & $FFFF ;*S2 LSW
;	                    PEA #S1_PTR >> 16   ;*S1 MSW
;	                    PEA #S1_PTR & $FFFF ;*S1 LSW
;	                    JSR STRCAT          ;concatenate S2 to S1
;	                    BCS ERROR           ;string too long
;
;	Exit registers: .A: entry value
;	                .B: entry value
;	                .X: entry value
;	                .Y: entry value
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: string too long
;	                    +++++++> entry value
;	
;
strcat   clc                   ;assume no error
;
;
;
;LOCAL DEFINTIONS
;
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> workspace stack frame start <
.s1siz   =.sfidx               ;S1's size
.sfidx   .= .sfidx+.s_word
;> workspace stack frame end <
;
.s_wsf   =.sfidx-.sfbase       ;workspace size
.sfbase  .= .sfidx
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;
;> parameter stack frame start <
.s1ptr   =.sfidx               ;*S1
.sfidx   .= .sfidx+.s_dword
.s2ptr   =.sfidx               ;*S2
.sfidx   .= .sfidx+.s_dword
;> parameter stack frame end <
;
.s_psf   =.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm    ;16-bit registers & binary mode
         phy                   ;save entry state
         phx
         pha
         phb
         phd
         sec
         tsc                   ;SP > .C
         sbc !#.s_wsf          ;allocate local workspace
         tcs                   ;.C > SP
         inc                   ;point DP to...
         tcd                   ;local workspace
	_strlen_ .s1ptr       ;get siz(S1)
         bmi .stlerr           ;string size error
;
         sty .s1siz            ;save siz(S1)
	_strlen_ .s2ptr       ;get siz(S2)
         bmi .stlerr           ;siz(S2) > $7FFF
;
         clc
         tya                   ;siz(S2)
         adc .s1siz            ;siz(S1)
         bmi .stlerr           ;siz(S1)+siz(S2) > $7FFF
;
;
;	compute where to catenate...
;
         clc
         lda .s1ptr            ;*S1 LSW
         adc .s1siz            ;siz(S1)
         sta .s1ptr            ;catenate here A0-A15
         sep #.seta            ;8-bit accumulator
         lda .s1ptr+.s_word    ;*S1 LSB of MSW
         adc #0
         bcs .stlerr           ;bank wrap  error
;
         sta .s1ptr+.s_word    ;catenate here A16-A23
         ldy !#0
;
;
;	catenate S2 to S1...
;
.main    lda [.s2ptr],y
         sta [.s1ptr],y
         beq .done             ;EOS
;
         iny
         bpl .main             ;next
;
         bra .stlerr           ;this shouldn't happen!
;
;
;	clean up stack & return to caller...
;
.done    rep #.setr|.sr_car    ;16-bit registers & clear carry
         tsc                   ;SP > .C
         adc !#.s_wsf          ;local workspace
         tcs                   ;get rid of it
         adc !#.s_rsf          ;register frame
         tax                   ;copy-from address
         adc !#.s_psf          ;parameter frame
         tay                   ;copy-to address
         lda !#.s_rsf-1        ;register frame size
         mvp #0,#0             ;move register frame, ...
         tya                   ;which wipes out...
         tcs                   ;parameter frame
         pld                   ;restore entry state
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
;
;	error handler  sets carry for return...
;
.stlerr  sep #.seta
         lda #.sr_car
         tsb .reg_sr
         bra .done
;
	.endif
	.if .ref(strchr)
;===============================================================================
;
;strchr: SEARCH STRING S FOR CHARACTER C
;
;	
;	Invocation example: PEA #C_PTR >> 16    ;*C MSW
;	                    PEA #C_PTR & $FFFF  ;*C LSW
;	                    PEA #S_PTR >> 16    ;*S pointer MSW
;	                    PEA #S_PTR & $FFFF  ;*S pointer LSW
;	                    JSR STRCHR          ;find C in S
;	                    BCS ERROR           ;string size error
;
;	                    BEQ NOTFOUND        ;C not in S
;
;	                    STX COUNT           ;instances of C
;	                    STY INDEX           ;index to 1st C instance
;
;	Exit registers: .A: entry value
;	                .B: entry value
;	                .X: instances of C in S  
;	                .Y: index to 1st instance of C  
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: string too long
;	                    ||||+++> entry value
;	                    |||+> 0 
;	                    +++> entry value
;
;	1) $0000 if C not found.
;	2) Entry value if string-too-long error
;	
;
strchr   clc                   ;assume no error
;
;
;
;LOCAL DEFINTIONS
;
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> workspace stack frame start <
.charidx =.sfidx               ;C's 1st location in S
.sfidx   .= .sfidx+.s_word
;> workspace stack frame end <
;
.s_wsf	=.sfidx-.sfbase       ;workspace frame size
.sfbase	.= .sfidx
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;
;> parameter stack frame start <
.sptr    =.sfidx               ;*S
.sfidx   .= .sfidx+.s_dword
.cptr    =.sfidx               ;*C
.sfidx   .= .sfidx+.s_dword
;> parameter stack frame end <
;
.s_psf   =.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm    ;16-bit registers & binary mode
         phy                   ;save entry state
         phx
         pha
         phb
         phd
         sec
         tsc
         sbc !#.s_wsf          ;allocate workspace
         tcs
         inc                   ;point DP to...
         tcd                   ;reserved stack space
         sep #.seta
         ldy !#0               ;S index
         tyx                   ;instances of C in S
         stx .charidx          ;reset C index in S
;
.main    lda [.sptr],y         ;get from S
         beq .eos              ;EOS, process results
;
         cmp [.cptr]           ;compare to C
         beq .main02           ;found instance
;
.main01  iny
         bpl .main             ;continue
;
         sep #.seta
         lda #.sr_car          ;string-too-long...
         tsb .reg_sr           ;error &...
         bra .done             ;abort
;
.main02  cpx !#0               ;1st instance of C?
         bne .main03           ;no
;
         sty .charidx          ;yes, set location index
;
.main03  inx                   ;bump instances count
         bra .main01           ;resume searching
;
.eos     lda #.setx
         ldy .charidx          ;index to 1st instance
         trb .reg_sr           ;16-bit index on return
         stx .reg_x            ;return instances
         sty .reg_y            ;return index
;
.done    rep #.setr|.sr_car
         tsc
         adc !#.s_wsf
         tcs
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
	.endif
	.if .ref(strcmp)
;===============================================================================
;
;strcmp: COMPARE STRING S2 TO STRING S1
;
;	
;	Invocation example: PEA #S2_PTR >> 16   ;*S2 MSW
;	                    PEA #S2_PTR & $FFFF ;*S2 LSW
;	                    PEA #S1_PTR >> 16   ;*S1 MSW
;	                    PEA #S1_PTR & $FFFF ;*S1 LSW
;	                    JSR STRCMP          ;compare strings
;	                    BCS ERROR           ;string too long
;
;	                    BEQ EQUAL           ;S2 = S1
;	                    BMI LESSER          ;S2 < S1
;	                    BPL GREATER         ;S2 > S1
;
;	     Registers: .A: entry value
;	                .B: entry value
;	                .X: entry value
;	                .Y: entry value
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: string too long
;	                    ||||||+> see invocation example 
;	                    |+++++> entry value
;	                    +> see invocation example 
;
;	1) Entry value if a string-too-long error occurs.
;	
;
strcmp   clc                   ;assume no error
;
;
;
;LOCAL DEFINTIONS
;
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> workspace stack frame start <
.s1siz   =.sfidx               ;S1's size
.sfidx   .= .sfidx+.s_word
.s2siz   =.sfidx               ;S2's size
.sfidx   .= .sfidx+.s_word
;> workspace stack frame end <
;
.s_wsf	=.sfidx-.sfbase       ;workspace frame size
.sfbase	.= .sfidx
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;
;> parameter stack frame start <
.s1ptr   =.sfidx               ;*S1
.sfidx   .= .sfidx+.s_dword
.s2ptr   =.sfidx               ;*S2
.sfidx   .= .sfidx+.s_dword
;> parameter stack frame end <
;
.s_psf   =.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm    ;16-bit registers & binary mode
         phy                   ;save entry state
         phx
         pha
         phb
         phd
         sec
         tsc
         sbc !#.s_wsf          ;allocate workspace
         tcs
         inc                   ;point DP to...
         tcd                   ;reserved stack space
         sep #.seta
         _strlen_ .s2ptr       ;get siz(S2)
         bmi .stlerr           ;string size error
;
         sty .s2siz            ;save siz(S2)
         _strlen_ .s1ptr       ;get siz(S1)
         bmi .stlerr           ;string size error
;
         sty .s1siz            ;save siz(S1)
         lda #.sr_neg|.sr_zer  ;reset...
         trb .reg_sr           ;result flags
         cpy .s2siz            ;siz(S2)
         beq .setup            ;siz(S1) = siz(S2)
;
         bcc .setup            ;siz(S1) < siz(S2)
;
         ldy .s2siz            ;siz(S1) > siz(S2)
;
.setup   tyx                   ;byte counter
         beq .main01           ;skip string scan
;
         ldy !#0
;
.main    lda [.s2ptr],y        ;S2
         cmp [.s1ptr],y        ;S1
         bne .main02           ;mismatch
;
         iny
         dex
         bne .main             ;next
;
.main01  ldx .s2siz            ;S2's size
         cpx .s1siz            ;S1's size
         bne .main02           ;S2 != S1
;
         lda #.sr_zer          ;S2 = S1
         bra .main03
;
.main02  bcs .done             ;S2 > S1
;
         lda #.sr_neg          ;S2 < S1
;
.main03  tsb .reg_sr           ;condition results
;
.done    rep #.setr|.sr_car    ;return to caller
         tsc
         adc !#.s_wsf
         tcs
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld                   ;restore entry state
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
.stlerr  lda #.sr_car
         tsb .reg_sr
         bra .done
;
	.endif
	.if .ref(strcpy)
;===============================================================================
;
;strcpy: COPY STRING S2 to STRING S1
;
;	
;	Invocation example: PEA #S2_PTR >> 16   ;*S2 MSW
;	                    PEA #S2_PTR & $FFFF ;*S2 LSW
;	                    PEA #S1_PTR >> 16   ;*S1 MSW
;	                    PEA #S1_PTR & $FFFF ;*S1 LSW
;	                    JSR STRCPY          ;copy S2 to S1
;	                    BCS ERROR           ;string too long
;
;	Exit registers: .A: entry value
;	                .B: entry value
;	                .X: entry value
;	                .Y: entry value
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: string too long
;	                    +++++++> entry value
;	
;
strcpy   clc                   ;assume no error
;
;
;
;LOCAL DEFINTIONS
;
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;
;> parameter stack frame start <
.s1ptr   =.sfidx               ;*S1
.sfidx   .= .sfidx+.s_dword
.s2ptr   =.sfidx               ;*S2
.sfidx   .= .sfidx+.s_dword
;> parameter stack frame end <
;
.s_psf   =.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm
         phy
         phx
         pha
         phb
         phd
         tsc
         inc
         tcd
         sep #.seta
         _strlen_ .s2ptr       ;get siz(S2)
         bmi .stlerr           ;string size error
;
         ldx .s2ptr+.s_word    ;*S2 MSW
         cpx .s1ptr+.s_word    ;*S1 MSW
         bne .setup            ;different, continue
;
         ldx .s2ptr            ;*S2 LSW
         cpx .s1ptr            ;*S1 LSW
         beq .done             ;same string, abort
;
.setup   ldy !#0
;
.main    lda [.s2ptr],y        ;overwrite S1...
         sta [.s1ptr],y        ;with S2
         beq .done
;
         iny
         bpl .main
;
.stlerr  lda #.sr_car
         tsb .reg_sr
;
.done    rep #.setr|.sr_car    ;return to caller
         tsc                   ;clean up stack
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
	.endif
	.if .ref(strdel)
;===============================================================================
;
;strdel: DELETE SUBSTRING OF N BYTES FROM STRING S AT INDEX I
;
;	
;	Invocation example: PEA #N_PTR >> 16    ;*N MSW
;	                    PEA #N_PTR & $FFFF  ;*N LSW
;	                    PEA #I_PTR >> 16    ;*I MSW
;	                    PEA #I_PTR & $FFFF  ;*I LSW
;	                    PEA #S_PTR >> 16    ;*S MSW
;	                    PEA #S_PTR & $FFFF  ;*S LSW
;	                    JSR STRDEL          ;delete substring
;	                    BCS ERROR           ;error
;
;	Exit registers: .A: entry value
;	                .B: entry value
;	                .X: entry value
;	                .Y: entry value
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: error
;	                    +++++++> entry value
;
;	
;
strdel   clc                   ;assume no error
;
;
;
;LOCAL DEFINTIONS
;
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> workspace stack frame start <
.dhptr	=.sfidx               ;shift destination pointer
.sfidx	.= .sfidx+.s_dword
.srptr	=.sfidx               ;shift source pointer
.sfidx	.= .sfidx+.s_dword
.ssiz	=.sfidx               ;S's old size (SS)
.sfidx	.= .sfidx+.s_word
.wsiz	=.sfidx               ;S's new size (WS)
.sfidx	.= .sfidx+.s_word
;> workspace stack frame end <
;
.s_wsf	=.sfidx-.sfbase       ;workspace frame size
.sfbase	.= .sfidx
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;
;> parameter stack frame start <
.sptr	=.sfidx               ;*S
.sfidx	.= .sfidx+.s_dword
.iptr	=.sfidx               ;*I
.sfidx	.= .sfidx+.s_dword
.nptr	=.sfidx               ;*N
.sfidx	.= .sfidx+.s_dword
;> parameter stack frame end <
;
.s_psf	=.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm
         phy
         phx
         pha
         phb
         phd
         sec
         tsc
         sbc !#.s_wsf          ;allocate workspace
         tcs
         inc                   ;point DP to...
         tcd                   ;reserved stack space
	_strlen_ .sptr        ;get siz(S)
         bmi .stlerr           ;too big
;
         lda [.nptr]           ;N
         beq .done             ;N = 0 (no operation)
;
         tya
         beq .done             ;SS = 0 (no operation)
;
         cmp [.iptr]           ;I
         bcc .done             ;SS < I (no operation)
;
         beq .done             ;SS = I (no operation)
;
         sty .ssiz             ;save SS
         lda [.iptr]           ;I
         sta .wsiz             ;set tentative WS
         clc
         adc [.nptr]           ;I+N
         cmp .ssiz             ;is I+N >= SS?
         bcs .main01           ;yes (no shift)
;
         sec                   ;no (must shift)
         tya                   ;SS
         sbc [.nptr]           ;SS-N
         sta .wsiz             ;set final WS
         clc
         lda .sptr             ;*S LSW
         adc [.iptr]           ;*S+I LSW
         sta .dhptr            ;shift destination LSW
         tax                   ;save a copy
         lda .sptr+.s_word     ;*S MSW
         adc !#0               ;*S+I MSW
         sta .dhptr+.s_word    ;shift destination MSW
         tay                   ;save a copy
         clc
         txa                   ;shift destination LSW
         adc [.nptr]           ;*S+I+N LSW
         sta .srptr            ;shift source LSW
         tya                   ;shift destination MSW
         adc !#0               ;*S+I+N MSW
         sta .srptr+.s_word    ;shift source MSW
         sec
         lda .ssiz             ;SS
         sbc [.iptr]           ;SS-I
         inc                   ;SS-I+1
         tax                   ;shift counter
         sep #.seta
         ldy !#0               ;index
;
.main    lda [.srptr],y        ;shift...
         sta [.dhptr],y        ;remainder
         iny
         dex
         bne .main
;
.main01  sep #.seta
         lda #0
         ldy .wsiz
         sta [.sptr],y         ;set new terminator

.done    rep #.setr|.sr_car    ;return to caller
         tsc                   ;clean up stack
         adc !#.s_wsf
         tcs
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld                   ;restore entry state
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
.stlerr  sep #.seta
         lda #.sr_car
         tsb .reg_sr
         rep #.seta
         bra .done
;
	.endif
	.if .ref(strins)
;===============================================================================
;
;strins: INSERT STRING S2 INTO STRING S1 AT INDEX I
;
;	
;	Invocation example: PEA #I_PTR >> 16    ;*I MSW
;	                    PEA #I_PTR & $FFFF  ;*I LSW
;	                    PEA #S2_PTR >> 16   ;*S2 MSW
;	                    PEA #S2_PTR & $FFFF ;*S2 LSW
;	                    PEA #S1_PTR >> 16   ;*S1 MSW
;	                    PEA #S1_PTR & $FFFF ;*S1 LSW
;	                    JSR STRINS          ;insert S2 into S1
;	                    BCS ERROR           ;string too long
;
;	Exit registers: .A: entry value
;	                .B: entry value
;	                .X: entry value
;	                .Y: entry value
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: string too long
;	                    +++++++> entry value
;	
;
strins   clc                   ;assume no error
;
;
;
;LOCAL DEFINTIONS
;
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> workspace stack frame start <
.s1siz   =.sfidx               ;S1's entry size
.sfidx   .= .sfidx+.s_word
.s2siz   =.sfidx               ;S2's size
.sfidx   .= .sfidx+.s_word
.fsiz    =.sfidx               ;S1's expanded size
.sfidx   .= .sfidx+.s_word
.w1ptr   =.sfidx               ;append/insert pointer
.sfidx   .= .sfidx+.s_dword
.w2ptr   =.sfidx               ;shift-to pointer
.sfidx   .= .sfidx+.s_dword
;> workspace stack frame end <
;
.s_wsf	=.sfidx-.sfbase       ;workspace frame size
.sfbase	.= .sfidx
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;> parameter stack frame start <
.s1ptr   =.sfidx               ;*S1
.sfidx   .= .sfidx+.s_dword
.s2ptr   =.sfidx               ;*S2
.sfidx   .= .sfidx+.s_dword
.iptr    =.sfidx               ;*I
.sfidx   .= .sfidx+.s_dword
;> parameter stack frame end <
;
.s_psf   =.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm
         phy
         phx
         pha
         phb
         phd
         sec
         tsc
         sbc !#.s_wsf          ;allocate workspace
         tcs
         inc                   ;point DP to...
         tcd                   ;reserved stack space
         ldx .s1ptr            ;*S1 LSW
         ldy .s1ptr+.s_word    ;*S1 MSW
         cpy .s2ptr+.s_word    ;*S2 MSW
         bne .setup            ;different strings, okay
;
         cpx .s2ptr            ;*S2 LSW
         bne .setup            ;different strings, okay
;
         brl .done             ;same string, abort
;
.setup   per .strlenr-1
	brl _strlen_          ;get siz(S1)
.strlenr bmi .stlerr           ;too big
;
         sty .s1siz            ;save siz(S1)
	_strlen_ .s2ptr       ;get siz(S2)
         bpl .setup01          ;ok
;
.stlerr  sep #.seta
         lda #.sr_car
         tsb .reg_sr
         bra .done
;
.setup01 clc
         tya                   ;siz(S2)
         adc .s1siz            ;siz(S2) + siz(S1)
         bmi .stlerr           ;too big
;
         sty .s2siz            ;save siz(S2)
         sta .fsiz             ;S1's expanded size
         lda .s1siz            ;old siz(S1)
         cmp [.iptr]           ;I
         bcc .done             ;I > siz(S1), no operation
;
         bne .setup02          ;I < siz(S1), insert
;
         clc                   ;I = siz(S1), append
         adc .s1ptr            ;*S1 LSW
         sta .w1ptr            ;append-at LSW
         lda .s1ptr+.s_word    ;*S1 MSW
         adc !#0
         sta .w1ptr+.s_word    ;append-at MSW
         sep #.seta
         bra .main01
;
.setup02 clc
         lda .s1ptr            ;*S1 LSW
         adc [.iptr]           ;I LSW
         sta .w1ptr            ;insert-at LSW
         tax                   ;make a copy
         lda .s1ptr+.s_word    ;*S1 MSW
         adc !#0               ;I "MSW"
         sta .w1ptr+.s_word    ;insert-at MSW
         tay                   ;make a copy
         txa                   ;insert-at LSW
         adc .s2siz            ;bytes to insert
         sta .w2ptr            ;shift-to LSW
         tya                   ;insert-at MSW
         adc !#0
         sta .w2ptr+.s_word    ;shift-to MSW
         sec
         lda .s1siz            ;siz(S1)
         sbc [.iptr]           ;I
         tay                   ;bytes to shift
         dey                   ;zero align
         sep #.seta
;
.main    lda [.w1ptr],y        ;shift S1 to make...
         sta [.w2ptr],y        ;room for S2
         dey
         bpl .main
;
.main01  ldy !#0
;
.main02  lda [.s2ptr],y        ;get from S2
         beq .main03           ;EOS
;
         sta [.w1ptr],y        ;append/insert to S1
         iny
         bpl .main02
;
         bra .stlerr           ;this shouldn't happen!
;
.main03  ldy .fsiz             ;S1's expanded size
         sta [.s1ptr],y        ;terminate S1
;
.done    rep #.setr|.sr_car    ;return to caller
         tsc                   ;clean up stack
         adc !#.s_wsf
         tcs
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
	.endif
	.if .ref(strlen)
;===============================================================================
;
;strlen: DETERMINE SIZE IN BYTES OF STRING S
;
;	
;	Invocation example: PEA #S_PTR >> 16    ;*S pointer MSW
;	                    PEA #S_PTR & $FFFF  ;*S pointer LSW
;	                    JSR STRLEN          ;get text field size
;	                    BCS ERROR           ;string too long
;
;	                    STA <size>          ;size returned in .C
;
;	Exit registers: .C: text field size 
;	                .X: entry value
;	                .Y: entry value
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: error
;	                    ||||||+> 0: size > 0 
;	                    ||||||       1: size = 0 
;	                    |||+++> entry value
;	                    ||+> 0 
;	                    ++> entry value
;
;	Entry value if an error occurs.
;	
;
strlen   clc                   ;assume no error
;
;
;
;LOCAL DEFINTIONS
;
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;
;> parameter stack frame start <
.sptr    =.sfidx               ;*S
.sfidx   .= .sfidx+.s_dword
;> parameter stack frame end <
;
.s_psf   =.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm
         phy
         phx
         pha
         phb
         phd
         tsc
         inc
         tcd
         sep #.seta
	_strlen_ .sptr        ;get siz(S)
         bmi .stlerr           ;string too big
;
         php                   ;save state of Z
         sty .reg_c            ;return size to caller
         lda #.seta|.sr_zer    ;m = 0 &...
         trb .reg_sr           ;Z = 0
         pla                   ;retrieve MPU status,...
         and #.sr_zer          ;isolate Z &...
         tsb .reg_sr           ;condition returned SR
;
.done    rep #.setr|.sr_car    ;return to caller
         tsc                   ;clean up stack
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld                   ;restore entry state
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
.stlerr  sep #.seta
         lda #.sr_car
         tsb .reg_sr
         bra .done
;
	.endif
	.if .ref(strpad)
;===============================================================================
;
;strpad: COPY, JUSTIFY & PAD STRING: strpad *S1,*S2,*L,J,FB
;
;	
;	This function copies string S2 to string S1, overwriting S1, & pads or
;	truncates S1 according to the values pointed to by *L, & passed as J &
;	FB.
;
;	 *S1, *S2 & *L are 32-bit, little-endian pointers.  The MSB of each
;	  pointer's MSW should be $00.
;
;	 J & FB are 16-bit integerssee below for details.
;
;	 Prior to calling this function the pointer stack frame must be as
;	  follows:
;
;	  SP+15 > FB
;	  SP+13 > J
;	  SP+9  > *L
;	  SP+5  > *S2
;	  SP+1  > *S1
;
;	  Stack housekeeping will be performed upon exit.
;
;	 *S1 & *S2 point to null-terminated character strings, with a maximum
;	  permissible size of 32,767 bytes, not including the terminator.  If
;	  *S2 & *S1 point to the same string, no processing will occur & no er-
;	  ror will be returned.  Behavior is undefined if *S1 points to any
;	  part of S2's text field or vice versa.
;
;	 *L points to a 16-bit integer that determines the final size of S1
;	  after processing.  The maximum permissible value for L is 32,767).
;	  L's value will affect processing in one of the following ways:
;
;	   L = $0000
;
;	    S1 will be null.
;
;	   L < size(S2)
;
;	    S1 will be a copy of S2, with S1 truncated to L bytes.  Copying is
;	    from left to right, which means truncation will drop some trailing
;	    bytes.
;
;	   L = size(S2)
;
;	    S1 will be an exact copy of S2.
;
;	   L > size(S2)
;
;	    S1 will be a copy of S2 that has been padded to L size.
;
;	 J determines the method of justification that will be applied to S1
;	  when it is padded.  Valid values for J are:
;
;	      J  Justification   Mnemonic
;	      -------------
;	      0      Left        _JUSTLFT
;	      1      Right       _JUSTRIT
;	      2      Center      _JUSTCTR
;	      -------------
;
;	  Any other value of J will be treated as 0 (left justification). The
;	  above mnemonics are defined in stringmacs.asm & should be used to
;	  define the value of J.
;
;	 FB defines the fill byte that will be used to pad S1 to the size def-
;	  ined by L when L > size(S2).  Any byte value other than $00 may be
;	  used as fill.  If FB = $00 an ASCII blank ($20) will be substituted.
;
;	 This function does not determine if the space allocated to S1 is suf-
;	  ficient to accommodate the string size defined by L.  It is the resp-
;	  onsibility of the calling function to allocate L+1 bytes to S1.
;
;	If the global symbol _STRLIB_ is defined to any value, assembly will be
;	such that this function must be called with JSL.  Otherwise, this func-
;	tion must be called with JSR.
;	
;	Invocation example: PEA #FB             ;FB (padding byte)
;	                    PEA #J              ;J (justification)
;	                    PEA #L_PTR >> 16    ;*L MSW
;	                    PEA #L_PTR & $FFFF  ;*L LSW
;	                    PEA #S2_PTR >> 16   ;*S2 MSW
;	                    PEA #S2_PTR & $FFFF ;*S2 LSW
;	                    PEA #S1_PTR >> 16   ;*S1 MSW
;	                    PEA #S1_PTR & $FFFF ;*S1 LSW
;	                    JSR STRPAD          ;copy & pad/truncate
;	                    BCS ERROR           ;string too long
;
;
;	Exit registers: .A: entry value
;	                .B: entry value
;	                .X: entry value
;	                .Y: entry value
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: error
;	                    +++++++> entry value
;
;	Returned for excessive string size or L out-of-range.
;	
;
strpad   clc                   ;assume no error
;
;
;LOCAL DEFINTIONS
;
.s_byte  =1                    ;size of byte
.s_word  =2                    ;size of word
.s_dword =4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb =.s_byte              ;data bank
.s_mpudp =.s_word              ;direct page
.s_mpupb =.s_byte              ;program bank
.s_mpupc =.s_word              ;program counter
.s_mpusp =.s_word              ;stack pointer
.s_mpusr =.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta   =%00100000             ;accumulator
.setx   =%00010000             ;index
.setr   =.seta | .setx         ;all
;
;
;	status register bits...
;
.sr_car  =%00000001            ;C  carry
.sr_bdm  =%00001000            ;D  decimal
.sr_irq  =%00000100            ;I  IRQ
.sr_neg  =%10000000            ;N  result negative
.sr_ovl  =%01000000            ;V  sign overflow
.sr_zer  =%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase  .= 0                  ;base stack index
.sfidx   .= .sfbase            ;workspace index
;
;> workspace stack frame start <
.w1ptr   =.sfidx               ;work pointer
.sfidx   .= .sfidx+.s_dword
.padsiz  =.sfidx               ;L
.sfidx   .= .sfidx+.s_word
.s2siz   =.sfidx               ;siz(S2)
.sfidx   .= .sfidx+.s_word
;> workspace stack frame end <
;
.s_wsf   =.sfidx-.sfbase       ;workspace size
.sfbase  .= .sfidx
;
;> register stack frame start <
.reg_dp  =.sfidx               ;DP
.sfidx   .= .sfidx+.s_mpudp
.reg_db  =.sfidx               ;DB
.sfidx   .= .sfidx+.s_mpudb
.reg_c   =.sfidx               ;.C
.sfidx   .= .sfidx+.s_word
.reg_x   =.sfidx               ;.X
.sfidx   .= .sfidx+.s_word
.reg_y   =.sfidx               ;.Y
.sfidx   .= .sfidx+.s_word
.reg_sr  =.sfidx               ;SR
.sfidx   .= .sfidx+.s_mpusr
.reg_pc  =.sfidx               ;PC
.sfidx   .= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb  =.sfidx               ;PB
.sfidx   .= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf   =.sfidx-.sfbase       ;register frame size
.sfbase  .= .sfidx
;
;> parameter stack frame start <
.s1ptr   =.sfidx               ;*S1
.sfidx   .= .sfidx+.s_dword
.s2ptr   =.sfidx               ;*S2
.sfidx   .= .sfidx+.s_dword
.lptr    =.sfidx               ;*L
.sfidx   .= .sfidx+.s_dword
.just    =.sfidx               ;J
.sfidx   .= .sfidx+.s_word
.padchr  =.sfidx               ;FB
.sfidx   .= .sfidx+.s_word
;> parameter stack frame end <
;
.s_psf   =.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm    ;16-bit registers & binary mode
         phy                   ;save entry state
         phx
         pha
         phb
         phd
         sec
         tsc
         sbc !#.s_wsf          ;allocate workspace
         tcs
         inc                   ;point DP to...
         tcd                   ;reserved stack space
         ldx .s2ptr            ;*S2 LSW
         ldy .s2ptr+.s_word    ;*S2 MSW
         cpy .s1ptr+.s_word    ;*S1 MSW
         bne .setup            ;different pointers, okay
;
         cpx .s1ptr            ;*S1 LSW
         beq .done             ;can't copy to self
;
.setup   per .strlenr-1
         brl __strlen__        ;get siz(S2)
.strlenr bmi .stlerr           ;string size error
;
         sty .s2siz            ;save siz(S2)
         ldx .s1ptr            ;*S1 LSW
         ldy .s1ptr+.s_word    ;*S1 MSW
         stx .w1ptr            ;work pointer LSW
         sty .w1ptr+.s_word    ;work pointer MSW
         lda [.lptr]           ;get L
         bpl .setup01          ;in range
;
.stlerr  sep #.seta            ;indicate a...
         lda #.sr_car          ;string-too-long...
         tsb .reg_sr           ;error
         bra .done             ;abort
;
.setup01 sta .padsiz           ;save L
         beq .main050          ;S1 will be null
;
         cmp .s2siz            ;L vs. siz(S2)
         bcc .main020          ;copy S2 w/truncation
;
         beq .main020          ;copy S2 verbatim
;
         tax                   ;copy S2 w/padding
         sep #.seta            ;8-bit accumulator
         lda .padchr           ;FB
         bne .setup02          ;use as is
;
         lda #' '              ;replace null w/blank
;
.setup02 ldy !#0               ;index
;
.main    sta [.w1ptr],y        ;fill S1 with...
         iny                   ;pad byte
         dex                   ;decrement counter
         bne .main
;
         ldx .just             ;J
         beq .main030          ;justify left
;
         rep #.seta            ;16-bit accumulator
         sec                   ;compute difference...
         lda .padsiz           ;between L &...
         sbc .s2siz            ;S2's size
         dex
         beq .main010          ;justify right
;
         dex
         bne .main020          ;justify left if J > 2
;
         lsr                   ;justify center
         adc !#0               ;center right if odd
;
.main010 clc
         adc .w1ptr            ;compute & set...
         sta .w1ptr            ;new work pointer LSW
         lda !#0
         adc .w1ptr+.s_word
         sta .w1ptr+.s_word    ;new work pointer MSW
;
.main020 sep #.seta            ;8-bit accumulator
;
.main030 ldx .padsiz           ;max text we should copy
         ldy !#0
;
.main040 lda [.s2ptr],y        ;get from S2
         beq .main060          ;end
;
         sta [.w1ptr],y        ;copy to S1
         iny
         dex                   ;keep going until we've...
         bne .main040          ;copied at most L chars
;
.main050 sep #.seta            ;8-bit accumulator
         lda #0                ;string terminator
;
.main060 ldy .padsiz           ;L
         sta [.s1ptr],y        ;terminate S1
;
.done    rep #.setr|.sr_car    ;return to caller
         tsc                   ;clean up stack
         adc !#.s_wsf
         tcs
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld                   ;restore entry state
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
             rtl
	.else
             rts
	.endif
;
	.endif
	.if .ref(strpat)
;===============================================================================
;
;strpat: PATTERN-MATCH STRING P TO STRING S USING WILDCARDS
;
;	
;	Invocation example: PEA #P_PTR >> 16    ;*P MSW
;	                    PEA #P_PTR & $FFFF  ;*P LSW
;	                    PEA #S_PTR >> 16    ;*S MSW
;	                    PEA #S_PTR & $FFFF  ;*S LSW
;	                    JSR STRPAT          ;pattern-match P against S
;	                    BCS ERROR           ;string too long
;
;	Exit registers: .A: entry value
;	                .B: entry value
;	                .X: entry value
;	                .Y: entry value
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: string too long
;	                    ||||||+> 0: P UNLIKE S 
;	                    ||||||       1: P LIKE S 
;	                    ++++++> entry value
;
;	1) 0 if an error occurs.
;	
;
strpat   ;this line intentionally has no code
;
;

;LOCAL DEFINTIONS
;
.cwild   ='?'                  ;character wildcard
.swild   ='*'                  ;string wildcard
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> workspace stack frame start <
.stptr	=.sfidx               ;S temp pointer
.sfidx	.= .sfidx+.s_dword
.ptptr	=.sfidx               ;P temp pointer
.sfidx	.= .sfidx+.s_dword
.swptr	=.sfidx               ;S work pointer
.sfidx	.= .sfidx+.s_dword
.pwptr	=.sfidx               ;P work pointer
.sfidx	.= .sfidx+.s_dword
;> workspace stack frame end <
;
.s_wsf	=.sfidx-.sfbase       ;workspace frame size
.sfbase	.= .sfidx
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;
;> parameter stack frame start <
.sptr    =.sfidx               ;*S
.sfidx   .= .sfidx+.s_dword
.pptr    =.sfidx               ;*P
.sfidx   .= .sfidx+.s_dword
;> parameter stack frame end <
;
.s_psf   =.sfidx-.sfbase       ;parameter stack frame size
;
;
         rep #.sr_zer|.sr_car  ;assume no error & UNLIKE
         php
         rep #.setr|.sr_bdm
         phy
         phx
         pha
         phb
         phd
         sec
         tsc
         sbc !#.s_wsf          ;allocate workspace
         tcs
         inc                   ;point DP to...
         tcd                   ;reserved stack space
         sep #.seta
         ldx .pptr             ;*P LSW
         ldy .pptr+.s_word     ;*P MSW
         stx .pwptr            ;set up work pointer LSW
         sty .pwptr+.s_word    ;set up work pointer MSW
         per .strlen1-1
         brl _strlen_          ;get siz(P)
.strlen1 bmi .stlerr           ;too big
;
         ldx .sptr             ;*S LSW
         ldy .sptr+.s_word     ;*S MSW
         stx .swptr            ;set up work pointer LSW
         sty .swptr+.s_word    ;set up work pointer MSW
         per .strlen2-1
         brl _strlen_          ;get siz(S)
.strlen2 bpl .main             ;size ok
;
.stlerr  sep #.seta
         lda #.sr_car
         tsb .reg_sr
         brl .done
;
.main    lda [.swptr]          ;get from S
         bne .main01
;
         brl .main13           ;end
;
.main01  lda [.pwptr]          ;get from P
         cmp #.swild           ;string wildcard?
         beq .main03           ;yes, process
;
         cmp [.swptr]          ;same as in S?
         beq .main02           ;yes, keep comparing
;
         cmp #.cwild           ;char wildcard?
         beq .main02           ;yes
;
         brl .main17           ;no, P UNLIKE S
;
.main02	per *+5
         brl .incp             ;*P = *P+1
         per *+5
         brl .incs             ;*S = *S+1
         bra .main             ;next comparison
;
.main03  lda [.swptr]          ;get from S
         beq .main13           ;EOS
;
         lda [.pwptr]          ;get from P
         cmp #.swild           ;string wildcard?
         bne .main07           ;no
;
         ldy !#1               ;get next...
         lda [.pwptr],y        ;in P
         beq .main16           ;EOS, P == S
;
         per *+5
         brl .incp             ;*P = *P+1
         ldx .pwptr            ;P work pointer LSW
         ldy .pwptr+.s_word    ;p work pointer MSW
         stx .ptptr            ;copy for...
         sty .ptptr+.s_word    ;position reset
         ldx .swptr            ;S work pointer LSW
         ldy .swptr+.s_word    ;S work pointer MSW
         inx                   ;point at next in S &...
         bne .main06           ;save in case...
;
         iny                   ;comparison fails
;
.main06  stx .stptr
         sty .stptr+.s_word
         lda [.pwptr]          ;get next from P
;
.main07  cmp [.swptr]          ;compare P to S
         beq .main08           ;match
;
         cmp #.cwild           ;char wildcard?
         bne .main11           ;no, mismatch
;
.main08  per *+5
         brl .incp             ;*P = *P+1
         per *+5
         brl .incs             ;*S = *S+1
         bra .main03           ;resume comparison
;
.main11  ldx .stptr            ;saved S work pointer LSW
         ldy .stptr+.s_word    ;saved S work pointer MSW
         stx .swptr            ;restore LSW
         sty .swptr+.s_word    ;restore MSW
         inx                   ;also point to next in S
         bne .main12
;
         iny
;
.main12  stx .stptr            ;point at next...
         sty .stptr+.s_word    ;in S for next compare
         ldx .ptptr            ;saved P work pointer LSW
         ldy .ptptr+.s_word    ;saved P work pointer MSW
         stx .pwptr            ;restore LSW
         sty .pwptr+.s_word    ;restore MSW
         bra .main03           ;resume comparison
;
.main13  ldy !#0
;
.main14  lda [.pwptr]          ;looking for trailing...
         cmp #.swild           ;string wildcards...
         bne .main15           ;in PATTERN
;
         iny
         bpl .main14           ;keep tring
;
.main15  ora #0                ;EOS?
         bne .main17           ;no, PATTERN != STRING
;
.main16  lda #.sr_zer          ;yes, PATTERN == STRING
         bra .done
;
.main17  lda #!.sr_zer         ;PATTERN != STRING
;
.done    ora .reg_sr           ;modify SR according...
         sta .reg_sr           ;to exit status
         rep #.setr|.sr_car    ;return to caller
         tsc                   ;clean up stack
         adc !#.s_wsf
         tcs
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
;	
;	increment *P
;	
;
.incp    rep #.seta
         inc .pwptr            ;bump LSW
         sep #.seta
         bne .0000020
;
         inc .pwptr+.s_word    ;bump LSB of MSW
         bne .0000020          ;okay
;
;
;	abort here if bank wraps...
;
.0000010 plx                   ;flush stack &...
         brl .stlerr           ;abort w/error
;
;	
;	increment *S
;	
;
.incs    rep #.seta
         inc .swptr            ;bump LSW
         sep #.seta
         bne .0000020
;
         inc .swptr+.s_word    ;bump LSB of MSW
         beq .0000010          ;oops!
;
.0000020 rts
;
	.endif
	.if .ref(strstr)
;===============================================================================
;
;strstr: FIND STRING S2 IN STRING S1
;
;	
;	Invocation example: PEA #S2_PTR >> 16   ;*S2 MSW
;	                    PEA #S2_PTR & $FFFF ;*S2 LSW
;	                    PEA #S1_PTR >> 16   ;*S1 MSW
;	                    PEA #S1_PTR & $FFFF ;*S1 LSW
;	                    JSR STRCAT          ;search S1 for S2
;	                    BCS ERROR           ;string too long
;
;	                    BEQ FOUND           ;S2 found in S1
;	                    BNE NOTFOUND        ;S2 not found in S1
;
;	Exit registers: .C: zero-based index to S2 in S1  
;	                .X: pointer to S2 in S1 LSW  
;	                .Y: pointer to S2 in S1 MSW  
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: error
;	                    ||||||+>  
;	                    ||||++> entry value
;	                    ||++> 0 
;	                    ++> entry value
;
;	Notes: 1) Entry value if string-too-long error.
;
;	       2) $0000 if S2 not found.
;
;	       3) 0 if no error & S2 was not found in S1.
;	          1 if no error & S2 was found in S1.
;	
;
strstr   clc                   ;assume no error
;
;
;
;LOCAL DEFINTIONS
;
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> workspace stack frame start <
.wptr	=.sfidx               ;S1 work pointer
.sfidx	.= .sfidx+.s_dword
.s1siz	=.sfidx               ;S1's size
.sfidx	.= .sfidx+.s_word
.s2siz	=.sfidx               ;S2's size
.sfidx	.= .sfidx+.s_word
;> workspace stack frame end <
;
.s_wsf	=.sfidx-.sfbase       ;workspace frame size
.sfbase	.= .sfidx
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;
;> parameter stack frame start <
.s1ptr	=.sfidx               ;*S1
.sfidx	.= .sfidx+.s_dword
.s2ptr	=.sfidx               ;*S1
.sfidx	.= .sfidx+.s_dword
;> parameter stack frame end <
;
.s_psf	=.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm
         phy
         phx
         pha
         phb
         phd
         sec
         tsc
         sbc !#.s_wsf          ;allocate workspace
         tcs
         inc                   ;point DP to...
         tcd                   ;reserved stack space
         stz .wptr             ;clear work pointer LSW
         stz .wptr+.s_word     ;clear work pointer MSW
         _strlen_ .s2ptr       ;get siz(S2)
         bmi .stlerr           ;S2 too big
;
         sty .s2siz            ;save siz(S2)
         _strlen_ .s1ptr       ;get siz(S1)
         bmi .stlerr           ;too big
;
         stz .reg_c            ;index = 0
         sep #.seta
         lda #.setr|.sr_zer    ;return w/16-bit registers &...
         trb .reg_sr           ;S2 not found as default
         cpy .s2siz            ;siz(S2)
         bcc .eos              ;siz(S2) > siz(S1), done
;
         ldy .s2siz            ;siz(S2)
         beq .eos              ;null, default is not found
;
         ldx .s1ptr            ;*S1 LSW
         ldy .s1ptr+.s_word    ;*S1 MSW
         stx .wptr             ;set work pointer LSW
         sty .wptr+.s_word     ;set work pointer MSW
;
.main    ldy !#0               ;string index
         sep #.seta
;
.main01  lda [.s2ptr],y        ;get from S2
         beq .eos              ;end, S2 found in S1
;
         cmp [.wptr],y         ;compare to S1
         bne .main02           ;not found so far
;
         iny
         bpl .main01           ;keep comparing
;
.stlerr  sep #.seta
         lda #.sr_car
         tsb .reg_sr
         bra .done
;
.main02  lda [.wptr],y         ;get from S1
         beq .main03           ;end, S2 not found
;
         rep #.seta
         inc .wptr             ;bump work pointer LSW
         bne .main             ;continue
;
         inc .wptr+.s_word     ;bump work pointer MSW
         bne .main             ;continue
;
.main03  ldx !#0
         stx .wptr             ;S2 not found,...
         stx .wptr+.s_word     ;null out pointer &...
;
.eos     rep #.seta
         lda .wptr             ;S2's location LSW
         sta .reg_x            ;return in .X
         lda .wptr+.s_word     ;S2's location MSW
         sta .reg_y            ;return in .Y
         ora .wptr             ;S2 found in S1?
         beq .done             ;no
;
         sec                   ;yes
         lda .wptr             ;S2's location LSW
         sbc .s1ptr            ;*S1
         sta .reg_c            ;index to S2 in S1
         sep #.seta
         lda #.sr_zer          ;set Z in SR to...
         tsb .reg_sr           ;indicate "found"
;
.done    rep #.setr|.sr_car    ;return to caller
         tsc                   ;clean up stack
         adc !#.s_wsf
         tcs
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
	.endif
	.if .ref(strsub)
;===============================================================================
;
;strsub: COPY N BYTES FROM STRING S2 INTO STRING S1 AT INDEX I
;
;	
;	Invocation example: PEA #N_PTR >> 16    ;*N MSW
;	                    PEA #N_PTR & $FFFF  ;*N LSW
;	                    PEA #I_PTR >> 16    ;*I MSW
;	                    PEA #I_PTR & $FFFF  ;*I LSW
;	                    PEA #S2_PTR >> 16   ;*S2 MSW
;	                    PEA #S2_PTR & $FFFF ;*S2 LSW
;	                    PEA #S1_PTR >> 16   ;*S1 MSW
;	                    PEA #S1_PTR & $FFFF ;*S1 LSW
;	                    JSR STRSUB          ;copy substring from S2 to S1
;	                    BCS ERROR           ;string too long
;
;	Exit registers: .A: entry value
;	                .B: entry value
;	                .X: entry value
;	                .Y: entry value
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: string too long
;	                    +++++++> entry value
;	
;
strsub   clc                   ;assume no error
;
;
;
;LOCAL DEFINTIONS
;
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;
;> parameter stack frame start <
.s1ptr	=.sfidx               ;*S1
.sfidx	.= .sfidx+.s_dword
.s2ptr	=.sfidx               ;*S2
.sfidx	.= .sfidx+.s_dword
.iptr	=.sfidx               ;*I pointer
.sfidx	.= .sfidx+.s_dword
.nptr	=.sfidx               ;*N pointer
.sfidx	.= .sfidx+.s_dword
;> parameter stack frame end <
;
.s_psf	=.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm
         phy
         phx
         pha
         phb
         phd
         sec
         tsc
         inc                   ;point DP to...
         tcd                   ;stack workspace
         ldx .s2ptr            ;*S2 LSW
         ldy .s2ptr+.s_word    ;*S2 MSW
         cpy .s1ptr+.s_word    ;*S1 MSW
         bne .setup            ;different strings
;
         cpx .s1ptr            ;*S1 LSW
         beq .done             ;same string, no operation
;
.setup   per .strlenr-1
         brl _strlen_          ;get siz(S2)
.strlenr bmi .stlerr           ;string size error
;
         tya
         cmp [.iptr]           ;I
         bcc .done             ;I > siz(S2)
;
         beq .done             ;i == siz(S2)
;
         lda [.nptr]           ;get N
         beq .done             ;$0000, no operation
;
         bmi .stlerr           ;>$7FFF  error
;
         tax                   ;N is counter
         clc
         lda .s2ptr            ;*S2 LSW
         adc [.iptr]           ;I
         bcc .setup01
;
         inc .s2ptr+.s_word    ;copy pointer MSW
         beq .stlerr           ;bank wrap  error
;
.setup01 sta .s2ptr            ;copy pointer LSW
         sep #.seta
         ldy !#0
;
.main    lda [.s2ptr],y        ;copy from S2
;
.main010 sta [.s1ptr],y
         beq .done             ;done copying
;
         iny                   ;bump index
         dex                   ;N = N-1
         bne .main             ;next
;
         txa                   ;$00 > .A
         beq .main010          ;terminate S1
;
.done    rep #.setr|.sr_car    ;return to caller
         tsc                   ;clean up stack
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
.stlerr  sep #.seta
         lda #.sr_car          ;string-too-long...
         tsb .reg_sr           ;error &...
         bra .done             ;abort
;
	.endif
	.if .ref(strsws)
;===============================================================================
;
;strsws: STRIP WHITESPACE FROM STRING
;
;	
;	Invocation example: PEA #OP             ;operation 
;	                    PEA #S_PTR >> 16    ;*S pointer MSW
;	                    PEA #S_PTR & $FFFF  ;*S pointer LSW
;	                    JSR STRSWS          ;strip whitespace from S
;	                    BCS ERROR           ;string size error
;
;	Exit registers: .A: entry value
;	                .B: entry value
;	                .X: entry value
;	                .Y: entry value
;	                DB: entry value
;	                DP: entry value
;	                PB: entry value
;	                SR: NVmxDIZC
;	                    ||||||||
;	                    |||||||+> 0: okay
;	                    |||||||      1: string too long
;	                    +++++++> entry value
;
;	1) OP: xx000000
;	       ||
;	       |+> 1: strip leading whitespace
;	       +> 1: strip trailing whitespace
;
;	   Operations may be combined.
;	
;
strsws   clc                   ;assume no error
;
;
;
;LOCAL DEFINTIONS
;
.a_htab  =$09                  ;horizontal tab
.a_blank =$20                  ;whitespace
.s_byte	=1                    ;size of byte
.s_word	=2                    ;size of word
.s_dword	=4                    ;size of long word
;
;
;	65C816 register sizes...
;
.s_mpudb	=.s_byte              ;data bank
.s_mpudp	=.s_word              ;direct page
.s_mpupb	=.s_byte              ;program bank
.s_mpupc	=.s_word              ;program counter
.s_mpusp	=.s_word              ;stack pointer
.s_mpusr	=.s_byte              ;status
;
;
;	65C816 register width masks...
;
.seta	=%00100000            ;accumulator
.setx	=%00010000            ;index
.setr	=.seta | .setx        ;all
;
;
;	status register bits...
;
.sr_car	=%00000001            ;C  carry
.sr_bdm	=%00001000            ;D  decimal
.sr_irq	=%00000100            ;I  IRQ
.sr_neg	=%10000000            ;N  result negative
.sr_ovl	=%01000000            ;V  sign overflow
.sr_zer	=%00000010            ;Z  result zero
;
;
;	stack definitions...
;
.sfbase	.= 0                  ;base stack index
.sfidx	.= .sfbase            ;workspace index
;
;> workspace stack frame start <
.wptr	=.sfidx               ;work pointer
.sfidx	.= .sfidx+.s_dword
;> workspace stack frame end <
;
.s_wsf	=.sfidx-.sfbase       ;workspace frame size
.sfbase	.= .sfidx
;
;> register stack frame start <
.reg_dp	=.sfidx               ;DP
.sfidx	.= .sfidx+.s_mpudp
.reg_db	=.sfidx               ;DB
.sfidx	.= .sfidx+.s_mpudb
.reg_c	=.sfidx               ;.C
.sfidx	.= .sfidx+.s_word
.reg_x	=.sfidx               ;.X
.sfidx	.= .sfidx+.s_word
.reg_y	=.sfidx               ;.Y
.sfidx	.= .sfidx+.s_word
.reg_sr	=.sfidx               ;SR
.sfidx	.= .sfidx+.s_mpusr
.reg_pc	=.sfidx               ;PC
.sfidx	.= .sfidx+.s_mpupc
	.if .ref(_STRLIB_)
.reg_pb	=.sfidx               ;PB
.sfidx	.= .sfidx+.s_mpupb
	.endif
;> register stack frame end <
;
.s_rsf	=.sfidx-.sfbase       ;register frame size
.sfbase	.= .sfidx
;
;> parameter stack frame start <
.sptr	=.sfidx               ;*S
.sfidx	.= .sfidx+.s_dword
.op	=.sfidx               ;OP
.sfidx	.= .sfidx+.s_word
;> parameter stack frame end <
;
.s_psf	=.sfidx-.sfbase       ;parameter frame size
;
;
         php
         rep #.setr|.sr_bdm    ;16-bit registers & binary mode
         phy                   ;save entry state
         phx
         pha
         phb
         phd
         sec
         tsc
         sbc !#.s_wsf          ;allocate workspace
         tcs
         inc                   ;point DP to...
         tcd                   ;reserved stack space
         sep #.seta
	_strlen_ .sptr        ;get siz(S)
         beq .done             ;null, just exit
;
         bmi .stlerr           ;siz(S) too big
;
         bit .op               ;check operation
         bpl .main020          ;no trailing strip
;
.main    dey                   ;decrement string index
         beq .main010          ;done w/trailing strip
;
         lda [.sptr],y         ;get from S
         cmp #.a_blank         ;blank?
         beq .main             ;yes, drop
;
         cmp #.a_htab          ;tab?
         beq .main             ;yes, drop
;
         iny                   ;move above whitespace
;
.main010 lda #0                ;mark new...
         sta [.sptr],y         ;end of S
         bit .op               ;recheck operation
;
.main020 bvc .done             ;no leading strip
;
         ldy !#0               ;prepare for leading strip
;
.main030 lda [.sptr],y         ;get from S
         beq .main050          ;EOS
;
         cmp #.a_blank         ;blank?
         beq .main040          ;yes, keep scanning
;
         cmp #.a_htab          ;horizontal tab?
         bne .main050          ;no
;
.main040 iny                   ;keep looking for...
         bpl .main030          ;whitespace
;
         bra .stlerr           ;this shouldn't happen!
;
.main050 rep #.seta
         clc
         tya                   ;current position in S
         adc .sptr             ;*S LSW
         sta .wptr             ;work pointer LSW
         lda !#0
         tay                   ;shift index
         adc .sptr+.s_word     ;*S MSW
         sta .wptr+.s_word     ;work pointer MSW
         sep #.seta
;
.main060 lda [.wptr],y         ;shift S to remove...
         sta [.sptr],y         ;leading whitespace
         beq .done             ;EOS, done
;
         iny
         bpl .main060
;
         bra .stlerr           ;this shouldn't happen!
;
.done    rep #.setr|.sr_car    ;16-bit registers & clear carry
         tsc                   ;clean up stack
         adc !#.s_wsf
         tcs
         adc !#.s_rsf
         tax
         adc !#.s_psf
         tay
         lda !#.s_rsf-1
         mvp #0,#0
         tya
         tcs
         pld                   ;restore entry state
         plb
         pla
         plx
         ply
         plp
	.if .ref(_STRLIB_)
         rtl
	.else
         rts
	.endif
;
.stlerr  sep #.seta
         lda #.sr_car
         tsb .reg_sr
         bra .done
;
	.endif
	.end
;
;This library is a collection of functions for manipulating ANSI or K&R C-style
;character strings, i.e., strings terminated by a null.  Usage of this library
;requires that the MPU be operated in native  mode.  Be sure to refer to the
;documentation for details on integrating this library into your application.

;This library version includes the following functions:

;  strcat  concatenate a string to a string
;  strchr  search a string for a byte
;  strcmp  compare two strings
;  strcpy  copy a string to a string
;  strdel  delete a substring from a string
;  strins  insert a substring into a string
;  strlen  get the size of a string
;  strpad  pad & justify a string
;  strpat  search a string for a pattern using wildcards
;  strstr  search a string for a string
;  strsub  copy a substring from a string
;  strsws  strip leading/trailing whitespace from a string

;Each function has a heading with an example of how to invoke the function as a
;subroutine call.  All functions expect parameters to be pushed on to the stack
;prior to making the function call with JSR or JSL  see the documentation.
;The invocation examples illustrate the process by using PEA.  Any method that
;places parameters of the correct size & order on the stack is acceptable.  PEI
;may be used if the parameters are on direct page, or one or more MPU registers
;may be employed.  Use the method that is most efficient for your particular
;application.

;This file will assemble in the Kowalski editor/assembler, version 1.3.4 or
;later.  Adapting this code to a different assembler should not be difficult.

;Notations used in comments:

;    WORD    16-bit, little-endian quantity
;    DWORD   32-bit, little-endian quantity
;    LSB     Least-significant byte
;    LSW     Least-significant word
;    MSB     Most-significant byte
;    MSW     Most-significant word
;    *X      Pointer to object X
;    X       Content of object X
;    siz(X)  Size of object X's content in bytes

;Pointers are always DWORDs.  As the 65C816 has 24 bits of addressing the MSB of
;a pointer's MSW is computationally irrelevent, but should be $00.
;
;This library makes no calls to other libraries.  Any function that makes a call
;to anything in this library must be placed in the source code so that said
;function is assembled before this libary.  If this requirement is not observed,
;a forward reference error will likely occur.
;